import unittest

from rdflib import ConjunctiveGraph, Graph
from rdflib import URIRef as U

from rdfdb.rdflibpatch import (contextsForStatement, graphFromNQuad, graphFromQuads, inContext, inGraph, patchQuads, serializeQuad)

A = U("http://a")
B = U("http://b")


class TestInContext(unittest.TestCase):

    def testResultHasQuads(self):
        g = inContext([(A, A, A)], B)
        self.assertEqual(list(g.quads())[0], (A, A, A, B))


class TestContextsForStatement(unittest.TestCase):

    def testNotFound(self):
        g = graphFromQuads([(A, A, A, A)])
        self.assertEqual(contextsForStatement(g, (B, B, B)), [])

    def testOneContext(self):
        g = graphFromQuads([(A, A, A, A), (A, A, B, B)])
        self.assertEqual(contextsForStatement(g, (A, A, A)), [A])

    def testTwoContexts(self):
        g = graphFromQuads([(A, A, A, A), (A, A, A, B)])
        self.assertEqual(sorted(contextsForStatement(g, (A, A, A))), sorted([A, B]))

    # There's a case where contextsForStatement was returning a Graph
    # with identifier, which I've fixed without a test


class TestInGraph(unittest.TestCase):

    def testSimpleMatch(self):
        g = graphFromQuads([(A, A, A, A)])
        self.assertTrue(inGraph((A, A, A, A), g))

    def testDontMatchDifferentStatement(self):
        g = graphFromQuads([(A, A, A, A)])
        self.assertFalse(inGraph((B, B, B, B), g))

    def testDontMatchStatementInAnotherContext(self):
        g = graphFromQuads([(A, A, A, A)])
        self.assertFalse(inGraph((A, A, A, B), g))

        self.assertFalse(inGraph((A, A, A, Graph(identifier=B)), g))


class TestGraphFromQuads(unittest.TestCase):
    nqOut = '<http://example.com/> <http://example.com/> <http://example.com/> <http://example.com/> .\n'

    def testSerializes(self):
        n = U("http://example.com/")
        g = graphFromQuads([(n, n, n, n)])
        out = serializeQuad(g)
        self.assertEqual(out.strip(), self.nqOut.strip())

    def testNquadParserSerializes(self):
        g = graphFromNQuad(self.nqOut)
        self.assertEqual(len(g), 1)
        out = serializeQuad(g)
        self.assertEqual(out.strip(), self.nqOut.strip())


A = U("http://a")
B = U("http://b")
C = U("http://c")
CTX1 = U('http://ctx1')
CTX2 = U('http://ctx2')
stmt1 = A, B, C, CTX1
stmt2 = A, B, C, CTX2


class TestPatchQuads(unittest.TestCase):

    def testAddsToNewContext(self):
        g = ConjunctiveGraph()
        patchQuads(g, [], [stmt1])
        self.assertEqual(len(g), 1)
        quads = list(g.quads((None, None, None)))
        self.assertEqual(quads, [(A, B, C, Graph(identifier=CTX1))])

    def testDeletes(self):
        g = ConjunctiveGraph()
        patchQuads(g, [], [stmt1])
        patchQuads(g, [stmt1], [])
        quads = list(g.quads((None, None, None)))
        self.assertEqual(quads, [])

    def testDeleteRunsBeforeAdd(self):
        g = ConjunctiveGraph()
        patchQuads(g, [stmt1], [stmt1])
        quads = list(g.quads((None, None, None)))
        self.assertEqual(quads, [(A, B, C, Graph(identifier=CTX1))])

    def testPerfectAddRejectsExistingStmt(self):
        g = ConjunctiveGraph()
        patchQuads(g, [], [stmt1])
        self.assertRaises(ValueError, patchQuads, g, [], [stmt1], perfect=True)

    def testPerfectAddAllowsExistingStmtInNewContext(self):
        g = ConjunctiveGraph()
        patchQuads(g, [], [stmt1])
        patchQuads(g, [], [stmt2], perfect=True)
        self.assertEqual(len(list(g.quads((None, None, None)))), 2)

    def testPerfectDeleteRejectsAbsentStmt(self):
        g = ConjunctiveGraph()
        self.assertRaises(ValueError, patchQuads, g, [stmt1], [], perfect=True)

    def testPerfectDeleteRejectsStmtFromOtherGraph(self):
        g = ConjunctiveGraph()
        patchQuads(g, [], [stmt2])
        self.assertRaises(ValueError, patchQuads, g, [stmt1], [], perfect=True)

    def testPerfectDeleteAllowsRemovalOfStmtInMultipleContexts(self):
        g = ConjunctiveGraph()
        patchQuads(g, [], [stmt1, stmt2])
        patchQuads(g, [stmt1], [], perfect=True)

    def testRedundantStmtOkForAddOrDelete(self):
        g = ConjunctiveGraph()
        patchQuads(g, [], [stmt1, stmt1], perfect=True)
        patchQuads(g, [stmt1, stmt1], [], perfect=True)
